#include "VINSController.hpp"

VINSController* VINSController::instance = nullptr;
ASensorEventQueue* VINSController::accelerometerEventQueue = nullptr;
ASensorEventQueue* VINSController::gyroscopeEventQueue = nullptr;

NSTimeInterval VINSController::systemUptime() {
    struct timespec time;
    clock_gettime(CLOCK_BOOTTIME, &time);
    return time.tv_sec + time.tv_nsec / 1000000.0f;
}

VINSController::VINSController() {
    LOGI("VINSController Constructor");
    this->instance = this;
}

VINSController::~VINSController() {
    LOGI("VINSController Destructor");
}

void VINSController::viewDidLoad() {
    isCapturing = false;
    if (!feature_tracker)
        feature_tracker = new FeatureTracker();

    // Check device to configure Intrinsic and Extrinsic
    bool deviceCheck = setGlobalParam(deviceName());
    if (!deviceCheck)
        LOGE("Device not supported.");
    vins.setExtrinsic();
    vins.setIMUModel();
    featuretracker.vins_pnp.setExtrinsic();
    featuretracker.vins_pnp.setIMUModel();

    mainLoop = std::thread(&VINSController::run, this);
    LOGI("THREAD: VINSController::run: thread created.");
    //saveDataThread = std::thread(&VINSController::saveDataLoop, this);

    if(LOOP_CLOSURE){
        loopDetectThread = std::thread(&VINSController::loopDetectionLoop, this);
        LOGI("THREAD: loopDetectThread created.");
        globalPoseOptmzThread = std::thread(&VINSController::globalPoseGraphOptimizationLoop, this);
        LOGI("THREAD: globalPoseOptmzThread created.");
    }

    imuStartUpdate();
    isCapturing = true;
    frameSize = cv::Size(videoWidth, videoHeight);
}

// Solve image feature and draw result view.
void VINSController::processImage(cv::Mat &image, double timeStamp, bool isScreenRotated) {

    if(isCapturing) {
        shared_ptr<IMG_MSG> img_msg(new IMG_MSG());
        img_msg->header = timeStamp;

        if(lateast_imu_time <= 0) {
            LOGD("IMU Timestamp negative: %lf, abort processImage()", lateast_imu_time);
            if(isScreenRotated)
                cv::rotate(image, image, cv::ROTATE_180);
            return;
        }

        assert(image.size() == frameSize);

        cv::Mat input_frame;
        input_frame = image;

        if(!imgDataBuf.empty()){
            LOGD("!imgDataBuf.empty(), abort processImage()");
            return;
        }

        cv::Mat gray;
        cv::cvtColor(input_frame, gray, CV_RGBA2GRAY);
        cv::Mat img_with_feature;
        cv::Mat img_equa, img_blur;

        // Histogram equalization
        double target_cliplimit;
        double g = cv::mean(gray)[0];
        if (g > 120){ target_cliplimit = 3; }
        else if (g > 60){
            target_cliplimit = 1.5 + (g - 60)/60 * 1.5;
        }
        else{ target_cliplimit = 1.5;}

        if (vins.solver_flag == VINS::NON_LINEAR){
            if (cur_cliplimit != target_cliplimit){
                cur_cliplimit += 0.1 * (target_cliplimit > cur_cliplimit ? 1 : -1 );
            }
        }
        else { cur_cliplimit = target_cliplimit;}

        cv::Ptr<cv::CLAHE> clahe = cv::createCLAHE();
        clahe->setClipLimit(cur_cliplimit);
        clahe->apply(gray, img_equa);

        // 3*3 medianBlur
        medianBlur(img_equa, img_blur, 3);

        // Featuretracker working, solves vins PnP
        m_depth_feedback.lock();
        featuretracker.solved_features = solved_features;
        featuretracker.solved_vins = solved_vins;
        m_depth_feedback.unlock();

        m_imu_feedback.lock();
        featuretracker.imu_msgs = getImuMeasurements(img_msg->header);
        m_imu_feedback.unlock();

        vector<Point2f> good_pts;
        vector<double> track_len;
        featuretracker.use_pnp = USE_PNP;
        featuretracker.readImage(img_equa, img_with_feature, frame_cnt, good_pts, track_len, img_msg->header, pnp_P, pnp_R, vins.solver_flag == VINS::NON_LINEAR);


        // Paint the points for the feature visualization
        for (int i = 0; i < good_pts.size(); i++) {
            cv::circle(image, good_pts[i], 0, cv::Scalar(255 * (1 - track_len[i]), 0, 255 * track_len[i]), 5); //BGR
        }

        // Push feature msg (detected by featuretracker) to img_msg_buf and notifying the main thread
        if(featuretracker.img_cnt == 0) {
            img_msg->point_clouds = featuretracker.image_msg;
            m_buf.lock();
            img_msg_buf.push(img_msg);
            m_buf.unlock();
            con.notify_one();
            if(imageCacheEnabled) {
                image_data_cache.header = img_msg->header;
                image_data_cache.image = image.clone();
                image_pool.push(image_data_cache);
            }

            if(LOOP_CLOSURE) {
                m_image_buf_loop.lock();
                cv::Mat loop_image = gray.clone();
                image_buf_loop.push(make_pair(loop_image, img_msg->header));
                if(image_buf_loop.size() > WINDOW_SIZE)
                    image_buf_loop.pop();
                m_image_buf_loop.unlock();
            }
        }
        featuretracker.img_cnt = (featuretracker.img_cnt + 1) % FREQ;

        if(imageCacheEnabled) {
            if(!vins_pool.empty() && !image_pool.empty()) {
                while(vins_pool.size() > 1) {
                    vins_pool.pop();
                }
                while(!image_pool.empty() && image_pool.front().header < vins_pool.front().header) {
                    image_pool.pop();
                }
                if(!vins_pool.empty() && !image_pool.empty()) {
                    lateast_image = image_pool.front().image;
                    lateast_P = vins_pool.front().P;
                    lateast_R = vins_pool.front().R;
                    image = lateast_image.clone();
                }
            }
            else if(!image_pool.empty()) {
                if(image_pool.size() > 10)
                    image_pool.pop();
            }
        }

        if(USE_PNP) {
            lateast_P = pnp_P.cast<float>();
            lateast_R = pnp_R.cast<float>();
        }
        if(ui_main || !start_show || vins.solver_flag != VINS::NON_LINEAR) {
            if(start_show && vins.solver_flag == VINS::NON_LINEAR) {
                cv::Mat tmp;
                vins.drawresult.startInit = true;
                vins.drawresult.drawAR(vins.imageAI, vins.correct_point_cloud, lateast_P, lateast_R);

                cv::cvtColor(image, tmp, CV_RGBA2RGB);
                cv::Mat mask;
                cv::Mat imageAI = vins.imageAI;
                if(!imageAI.empty())
                    cv::cvtColor(imageAI, mask, CV_RGB2GRAY);
                imageAI.copyTo(tmp,mask);
                                
                cv::cvtColor(tmp, image, CV_RGB2RGBA);
                if(isScreenRotated)
                    cv::rotate(image, image, cv::ROTATE_180);
            }
            else{
                if(isScreenRotated)
                    cv::rotate(image, image, cv::ROTATE_180);
            }
        }
        else  {
            // Draw VINS
            assert(vins.solver_flag == VINS::NON_LINEAR);
            if (vins.solver_flag == VINS::NON_LINEAR) {
                vins.drawresult.pose.clear();
                vins.drawresult.pose = keyframe_database.refine_path;
                vins.drawresult.segment_indexs = keyframe_database.segment_indexs;
                // Follow Camera
                vins.drawresult.origin_w[0] = (float) vins.correct_Ps[frame_cnt][0];
                vins.drawresult.origin_w[1] = (float) vins.correct_Ps[frame_cnt][1];
                vins.drawresult.radius = virtualCamDistance;
                // Rotation:
                vins.drawresult.theta = 75;
                vins.drawresult.phy = 89;
                vins.drawresult.Reprojection(vins.image_show, vins.correct_point_cloud,
                                             vins.correct_Rs, vins.correct_Ps, box_in_trajectory);
            }
            cv::Mat tmp2 = vins.image_show;

            cv::Mat down_origin_image;
            cv::resize(image, down_origin_image, cv::Size(150, 200));
            cv::rotate(down_origin_image, down_origin_image, cv::ROTATE_90_CLOCKWISE);
            cv::cvtColor(down_origin_image, down_origin_image, CV_RGBA2RGB);
            if (!isScreenRotated)
                cv::rotate(down_origin_image, down_origin_image, cv::ROTATE_180);

            cv::Mat imageROI;
            imageROI = tmp2(cv::Rect(0,COL - down_origin_image.rows, down_origin_image.cols,down_origin_image.rows));
            down_origin_image.copyTo(imageROI);

            cv::rotate(tmp2, image, cv::ROTATE_90_CLOCKWISE);
            cv::cvtColor(image, image, CV_RGB2RGBA);
        }
    }
    else {
        LOGD("Not capturing");
        if(isScreenRotated)
            cv::rotate(image, image, cv::ROTATE_180);
    }
}

std::vector<std::pair<std::vector<ImuConstPtr>, ImgConstPtr>> VINSController::getMeasurements() {
    std::vector<std::pair<std::vector<ImuConstPtr>, ImgConstPtr>> measurements;
    while (true) {
        if (imu_msg_buf.empty() || img_msg_buf.empty())
            return measurements;

        if (!(imu_msg_buf.back()->header > img_msg_buf.front()->header)) {
            __android_log_print(ANDROID_LOG_INFO, "VINS", "wait for imu, only should happen at the beginning");
            return measurements;
        }

        if (!(imu_msg_buf.front()->header < img_msg_buf.front()->header)) {
            __android_log_print(ANDROID_LOG_INFO, "VINS", "throw img, only should happen at the beginning");
            img_msg_buf.pop();
            continue;
        }
        ImgConstPtr img_msg = img_msg_buf.front();

        img_msg_buf.pop();

        std::vector<ImuConstPtr> IMUs;
        while (imu_msg_buf.front()->header <= img_msg->header) {
            IMUs.emplace_back(imu_msg_buf.front());
            imu_msg_buf.pop();
        }

        measurements.emplace_back(IMUs, img_msg);
    }
    return measurements;
}

vector<IMU_MSG_LOCAL> VINSController::getImuMeasurements(double header) {
    vector<IMU_MSG_LOCAL> imu_measurements;
    static double last_header = -1; // STATIC only happens once
    if(last_header < 0 || local_imu_msg_buf.empty()) {
        last_header = header;
        return imu_measurements;
    }

    while(!local_imu_msg_buf.empty() && local_imu_msg_buf.front().header <= last_header)
        local_imu_msg_buf.pop();
    while(!local_imu_msg_buf.empty() && local_imu_msg_buf.front().header <= header) {
        imu_measurements.emplace_back(local_imu_msg_buf.front());
        local_imu_msg_buf.pop();
    }
    last_header = header;
    return imu_measurements;
}

void VINSController::send_imu(const ImuConstPtr &imu_msg) {
    NSTimeInterval t = imu_msg->header;
    if (current_time < 0)
        current_time = t;
    double dt = (t - current_time);
    current_time = t;

    double ba[]{0.0, 0.0, 0.0};
    double bg[]{0.0, 0.0, 0.0};

    double dx = imu_msg->acc.x() - ba[0];
    double dy = imu_msg->acc.y() - ba[1];
    double dz = imu_msg->acc.z() - ba[2];

    double rx = imu_msg->gyr.x() - bg[0];
    double ry = imu_msg->gyr.y() - bg[1];
    double rz = imu_msg->gyr.z() - bg[2];

    vins.processIMU(dt, Vector3d(dx, dy, dz), Vector3d(rx, ry, rz));
}

void VINSController::run() {
    _condition.lock();
    while(!mainLoop_isCancelled) {
        process();
        std::this_thread::sleep_for(std::chrono::milliseconds(10));
    }
    _condition.unlock();

}

void VINSController::process() {
    std::vector<std::pair<std::vector<ImuConstPtr>, ImgConstPtr>> measurements;
    std::unique_lock<std::mutex> lk(m_buf);
    con.wait(lk, [&] {
        measurements = getMeasurements();
        return !measurements.empty();
    });
    lk.unlock();
    waiting_lists = measurements.size();
    for(auto &measurement : measurements) {
        TS(IMU_Data);
        for(auto &imu_msg : measurement.first){
            send_imu(imu_msg);
        }
        TE(IMU_Data);

        auto img_msg = measurement.second;
        map<int, Vector3d> image = img_msg->point_clouds;
        double header = img_msg->header;
        TS(process_image);
        vins.processImage(image, header, waiting_lists);
        TE(process_image);
        double time_now = lateast_imu_time;
        double time_vins = vins.Headers[WINDOW_SIZE];
        __android_log_print(ANDROID_LOG_DEBUG, "process()", "VINS delay %lf", time_now - time_vins);

        if(vins.solver_flag == vins.NON_LINEAR) {
            m_depth_feedback.lock();
            solved_vins.header = vins.Headers[WINDOW_SIZE - 1];
            solved_vins.Ba = vins.Bas[WINDOW_SIZE - 1];
            solved_vins.Bg = vins.Bgs[WINDOW_SIZE - 1];
            solved_vins.P = vins.correct_Ps[WINDOW_SIZE-1].cast<double>();
            solved_vins.R = vins.correct_Rs[WINDOW_SIZE-1].cast<double>();
            solved_vins.V = vins.Vs[WINDOW_SIZE - 1];
            Vector3d R_ypr = Utility::R2ypr(solved_vins.R);
            solved_features.clear();
            for (auto &it_per_id : vins.f_manager.feature) {
                it_per_id.used_num = it_per_id.feature_per_frame.size();
                if (!(it_per_id.used_num >= 2 && it_per_id.start_frame < WINDOW_SIZE - 2))
                    continue;
                if (it_per_id.solve_flag != 1)
                    continue;
                int imu_i = it_per_id.start_frame;
                Vector3d pts_i = it_per_id.feature_per_frame[0].point * it_per_id.estimated_depth;
                IMG_MSG_LOCAL tmp_feature;
                tmp_feature.id = it_per_id.feature_id;
                tmp_feature.position = vins.r_drift * vins.Rs[imu_i] * (vins.ric * pts_i + vins.tic) + vins.r_drift * vins.Ps[imu_i] + vins.t_drift;
                tmp_feature.track_num = (int)it_per_id.feature_per_frame.size();
                solved_features.push_back(tmp_feature);
            }
            m_depth_feedback.unlock();
        }

        if(imageCacheEnabled) {
            if(vins.solver_flag == VINS::NON_LINEAR && start_show) {
                VINS_DATA_CACHE vins_data_cache;
                vins_data_cache.header = vins.Headers[WINDOW_SIZE-1];
                vins_data_cache.P = vins.correct_Ps[WINDOW_SIZE-1];
                vins_data_cache.R = vins.correct_Rs[WINDOW_SIZE-1];
                vins_pool.push(vins_data_cache);
            }
            else if(vins.failure_occur == true) {
                vins.drawresult.change_color = true;
                vins.drawresult.indexs.push_back(vins.drawresult.pose.size());
                segmentation_index++;
                keyframe_database.max_seg_index++;
                keyframe_database.cur_seg_index = keyframe_database.max_seg_index;

                while(!vins_pool.empty())
                    vins_pool.pop();
            }
        }
        if(LOOP_CLOSURE) {
            static bool first_frame = true;
            if(vins.solver_flag != vins.NON_LINEAR)
                first_frame = true;
            if(vins.marginalization_flag == vins.MARGIN_OLD && vins.solver_flag == vins.NON_LINEAR && !image_buf_loop.empty()) {
                first_frame = false;
                if(!first_frame && keyframe_freq % LOOP_FREQ == 0) {
                    keyframe_freq = 0;
                    Vector3d T_w_i = vins.Ps[WINDOW_SIZE - 2];
                    Matrix3d R_w_i = vins.Rs[WINDOW_SIZE - 2];
                    m_image_buf_loop.lock();
                    while(!image_buf_loop.empty() && image_buf_loop.front().second < vins.Headers[WINDOW_SIZE - 2]) {
                        image_buf_loop.pop();
                    }
                    if(vins.Headers[WINDOW_SIZE - 2] == image_buf_loop.front().second) {
                        //TODO: parameterize this file path
                        const char *pattern_file = "/storage/emulated/0/VINS/brief_pattern.yml";
                        KeyFrame* keyframe = new KeyFrame(vins.Headers[WINDOW_SIZE - 2], global_frame_cnt, T_w_i, R_w_i, image_buf_loop.front().first, pattern_file, keyframe_database.cur_seg_index);
                        keyframe->setExtrinsic(vins.tic, vins.ric);
                        keyframe->buildKeyFrameFeatures(vins);
                        keyframe_database.add(keyframe);

                        global_frame_cnt++;
                    }
                    m_image_buf_loop.unlock();

                }
                else {
                    first_frame = false;
                }
                for (int i = 0; i < WINDOW_SIZE; i++) {
                    if(vins.Headers[i] == vins.front_pose.header) {
                        KeyFrame* cur_kf = keyframe_database.getKeyframe(vins.front_pose.cur_index);
                        if (abs(vins.front_pose.relative_yaw) > 30.0 || vins.front_pose.relative_t.norm() > 10.0) {
                            printf("Wrong loop\n");
                            cur_kf->removeLoop();
                            break;
                        }
                        cur_kf->updateLoopConnection(vins.front_pose.relative_t,
                                                     vins.front_pose.relative_q,
                                                     vins.front_pose.relative_yaw);
                        break;
                    }
                }
                int search_cnt = 0;
                for(int i = 0; i < keyframe_database.size(); i++) {
                    search_cnt++;
                    KeyFrame* kf = keyframe_database.getLastKeyframe(i);
                    if(kf->header == vins.Headers[0]) {
                        kf->updateOriginPose(vins.Ps[0], vins.Rs[0]);
                        if (kf->has_loop) {
                            kf_global_index = kf->global_index;
                            start_global_optimization = true;
                        }
                        break;
                    }
                    else {
                        if(search_cnt > 2 * WINDOW_SIZE)
                            break;
                    }
                }
                keyframe_freq++;
            }
        }
        waiting_lists--;

        showInputView();
    }
}

void VINSController::loopDetectionLoop() {
    if(LOOP_CLOSURE && loop_closure == NULL) {
        __android_log_print(ANDROID_LOG_INFO, APPNAME, "loop start load voc");
        const char *voc_file = "/storage/emulated/0/VINS/brief_k10L6.bin";
        loop_closure = new LoopClosure(voc_file, COL, ROW);
        __android_log_print(ANDROID_LOG_INFO, APPNAME, "loop load voc finish");

        voc_init_ok = true;
    }
    while(!loop_thread_isCancelled) {
        if(!LOOP_CLOSURE) {
            std::this_thread::sleep_for(std::chrono::milliseconds(500));
            continue;
        }

        bool loop_succ = false;
        if (loop_check_cnt < global_frame_cnt) {
            KeyFrame* cur_kf = keyframe_database.getLastKeyframe();
            loop_check_cnt++;
            cur_kf->check_loop = 1;

            cv::Mat current_image;
            current_image = cur_kf->image;

            std::vector<cv::Point2f> measurements_old;
            std::vector<cv::Point2f> measurements_old_norm;
            std::vector<cv::Point2f> measurements_cur;
            std::vector<int> features_id;
            std::vector<cv::Point2f> measurements_cur_origin = cur_kf->measurements;

            vector<cv::Point2f> cur_pts;
            vector<cv::Point2f> old_pts;

            cur_kf->extractBrief(current_image);
            
            printf("loop extract %lu feature\n", cur_kf->keypoints.size());
            TS(Loop_Detection);
            loop_succ = loop_closure->startLoopClosure(cur_kf->keypoints, cur_kf->descriptors, cur_pts, old_pts, old_index);
            TE(Loop_Detection);
            if(loop_succ) {
                KeyFrame* old_kf = keyframe_database.getKeyframe(old_index);
                if (old_kf == NULL) {
                    printf("NO such %dth frame in keyframe_database\n", old_index);
                    assert(false);
                }
                printf("loop succ with %drd image\n", old_index);
                assert(old_index!=-1);

                Vector3d T_w_i_old;
                Matrix3d R_w_i_old;

                old_kf->getPose(T_w_i_old, R_w_i_old);
                cur_kf->findConnectionWithOldFrame(old_kf, cur_pts, old_pts,
                                                   measurements_old, measurements_old_norm);
                measurements_cur = cur_kf->measurements;
                features_id = cur_kf->features_id;

                if(measurements_old_norm.size()>MIN_LOOP_NUM) {
                    Quaterniond Q_loop_old(R_w_i_old);
                    RetriveData retrive_data;
                    retrive_data.cur_index = cur_kf->global_index;
                    retrive_data.header = cur_kf->header;
                    retrive_data.P_old = T_w_i_old;
                    retrive_data.Q_old = Q_loop_old;
                    retrive_data.use = true;
                    retrive_data.measurements = measurements_old_norm;
                    retrive_data.features_ids = features_id;
                    vins.retrive_pose_data = (retrive_data);

                    cur_kf->detectLoop(old_index);
                    keyframe_database.addLoop(old_index);
                    old_kf->is_looped = 1;
                    loop_old_index = old_index;
                }
            }
            cur_kf->image.release();
        }

        if(loop_succ)
            std::this_thread::sleep_for(std::chrono::seconds(2));
        std::this_thread::sleep_for(std::chrono::milliseconds(50));
    }
}

void VINSController::globalPoseGraphOptimizationLoop() {
    while(!globalLoopThread_isCancelled)
    {
        if(start_global_optimization)
        {
            start_global_optimization = false;
            keyframe_database.optimize4DoFLoopPoseGraph(kf_global_index,
                                                        loop_correct_t,
                                                        loop_correct_r);
            vins.t_drift = loop_correct_t;
            vins.r_drift = loop_correct_r;
            std::this_thread::sleep_for(std::chrono::milliseconds(1170));
        }
        std::this_thread::sleep_for(std::chrono::milliseconds(30));
    }
}

void VINSController::imuStopUpdate() {
    ASensorManager *sensorManager = ASensorManager_getInstanceForPackage("com.bryan.tan.vins_ar");
    assert(sensorManager != NULL);
    const ASensor *accelerometer = ASensorManager_getDefaultSensor(sensorManager, ASENSOR_TYPE_ACCELEROMETER);
    const ASensor *gyroscope = ASensorManager_getDefaultSensor(sensorManager, ASENSOR_TYPE_GYROSCOPE);
    
    ASensorEventQueue_disableSensor(accelerometerEventQueue, accelerometer);
    ASensorEventQueue_disableSensor(gyroscopeEventQueue, gyroscope);
}

void VINSController::imuStartUpdate() {

    ASensorManager *sensorManager = ASensorManager_getInstanceForPackage("com.bryan.tan.vins_ar");
    assert(sensorManager != NULL);

    ALooper* looper = ALooper_forThread();
    if(looper == NULL) looper = ALooper_prepare(ALOOPER_PREPARE_ALLOW_NON_CALLBACKS);
    assert(looper != NULL);

    /** ACCELEROMETER PART */
    // E/Looper: Invalid attempt to set NULL callback but not allowed for this looper.
    // No callback but processed with gyro (process_imu_sensor_events()).
    accelerometerEventQueue = ASensorManager_createEventQueue(sensorManager, looper,
                                                              LOOPER_ID_USER, NULL,
                                                              NULL);
    assert(accelerometerEventQueue != NULL);
    const ASensor *accelerometer = ASensorManager_getDefaultSensor(sensorManager, ASENSOR_TYPE_ACCELEROMETER);
    assert(accelerometer != NULL);
    LOGI("ACCELEROMETER Sensor: %s", ASensor_getName(accelerometer));
    auto status = ASensorEventQueue_enableSensor(accelerometerEventQueue,accelerometer);
    assert(status >= 0);
    status = ASensorEventQueue_setEventRate(accelerometerEventQueue,accelerometer,SENSOR_REFRESH_PERIOD_US);
    assert(status >= 0);

    /** GYROSCOPE PART */   // Callback: process_imu_sensor_events
    gyroscopeEventQueue = ASensorManager_createEventQueue(sensorManager, looper,
                                                          LOOPER_ID_USER, process_imu_sensor_events,
                                                          NULL);
    assert(gyroscopeEventQueue != NULL);
    const ASensor *gyroscope = ASensorManager_getDefaultSensor(sensorManager, ASENSOR_TYPE_GYROSCOPE);
    assert(gyroscope != NULL);
    LOGI("GYROSCOPE Sensor: %s", ASensor_getName(gyroscope));
    status = ASensorEventQueue_enableSensor(gyroscopeEventQueue, gyroscope);
    assert(status >= 0);
    status = ASensorEventQueue_setEventRate(gyroscopeEventQueue, gyroscope, SENSOR_REFRESH_PERIOD_US);
    assert(status >= 0);

    LOGI("IMU EventQueues initialized and started");
}

int VINSController::process_imu_sensor_events(int fd, int events, void *data) {
    static ASensorEvent acclEvent;
    static double acclEventTimestamp = -1.0;
    ASensorEvent gyroEvent;

    while (ASensorEventQueue_getEvents(gyroscopeEventQueue, &gyroEvent, 1) > 0) {
        assert(gyroEvent.type == ASENSOR_TYPE_GYROSCOPE);

        double timeStampGyro = timeStampToSec(gyroEvent.timestamp);
        assert(timeStampGyro > 0);

        IMU_MSG gyro_msg;
        gyro_msg.header = timeStampGyro;
        gyro_msg.gyr << gyroEvent.uncalibrated_gyro.x_uncalib,
                gyroEvent.uncalibrated_gyro.y_uncalib,
                gyroEvent.uncalibrated_gyro.z_uncalib;

        if(instance->gyro_buf.size() == 0) {
            LOGI("gyro interpolation buffer empty. should only happen once.");
            instance->gyro_buf.push_back(gyro_msg);
            instance->gyro_buf.push_back(gyro_msg);
            continue;
        }
        else if(gyro_msg.header <= instance->gyro_buf[1].header) {
            continue;
        }
        else {
            instance->gyro_buf[0] = instance->gyro_buf[1];
            instance->gyro_buf[1] = gyro_msg;
        }

        if(instance->imu_prepare < 10) {
            instance->imu_prepare++;
            continue;
        }

        while (acclEventTimestamp < instance->gyro_buf[0].header) {
            ssize_t numEvents;
            while((numEvents = ASensorEventQueue_getEvents(accelerometerEventQueue, &acclEvent, 1)) == 0) {
                std::this_thread::sleep_for(std::chrono::milliseconds(1));
            }
            assert(numEvents == 1);
            assert(acclEvent.type == ASENSOR_TYPE_ACCELEROMETER);

            acclEventTimestamp = timeStampToSec(acclEvent.timestamp);

            shared_ptr<IMU_MSG> acc_msg(new IMU_MSG());
            acc_msg->header = acclEventTimestamp;
            acc_msg->acc << acclEvent.acceleration.x,
                    acclEvent.acceleration.y,
                    acclEvent.acceleration.z;
            instance->cur_acc = acc_msg;
        }
        if(instance->gyro_buf[1].header < acclEventTimestamp){
            LOGE("having to wait for fitting gyro event");
            continue;
        }

        // Interpolation (Linear)
        shared_ptr<IMU_MSG> imu_msg(new IMU_MSG());
        if(instance->cur_acc->header >= instance->gyro_buf[0].header &&
           instance->cur_acc->header < instance->gyro_buf[1].header) {
            imu_msg->header = instance->cur_acc->header;
            imu_msg->acc = instance->cur_acc->acc;
            imu_msg->gyr = instance->gyro_buf[0].gyr +
                           (instance->gyro_buf[1].gyr - instance->gyro_buf[0].gyr) *
                           (instance->cur_acc->header - instance->gyro_buf[0].header) /
                           (instance->gyro_buf[1].header - instance->gyro_buf[0].header);
        }
        else {
            LOGE("imu error %lf %lf %lf\n", instance->gyro_buf[0].header, instance->cur_acc->header, instance->gyro_buf[1].header);
            continue;
        }

        instance->lateast_imu_time = imu_msg->header;

        {
            IMU_MSG_LOCAL imu_msg_local;
            imu_msg_local.header = imu_msg->header;
            imu_msg_local.acc = imu_msg->acc;
            imu_msg_local.gyr = imu_msg->gyr;

            instance->m_imu_feedback.lock();
            instance->local_imu_msg_buf.push(imu_msg_local);
            instance->m_imu_feedback.unlock();
        }
        instance->m_buf.lock();
        instance->imu_msg_buf.push(imu_msg);
        instance->m_buf.unlock();
        instance->con.notify_one();
    }

    return 1;
}

void VINSController::saveDataLoop() {
    while(!saveData_isCancelled) {
        if(!imgDataBuf.empty()) {
            LOGI("saveDataLoop: Trying to save something");
            IMG_DATA tmp_data;
            tmp_data = imgDataBuf.front();
            imgDataBuf.pop();
            recordImageTime(tmp_data);
            recordImage(tmp_data);
            imageDataIndex++;
        }
        std::this_thread::sleep_for(std::chrono::milliseconds(40));
    }
}

DeviceType VINSController::deviceName() {
    return DeviceType::GalaxyS10e;
    //return DeviceType::LGV60;
}

void VINSController::showInputView() {
    std::ostringstream oss;

    static bool finish_init = false;

    viewUpdateMutex.lock();
    if (vins.solver_flag != vins.NON_LINEAR) {
        finish_init = false;
        switch (vins.init_status) {
            case VINS::InitStatus::FAIL_IMU:
                oss << "STA: FAIL_IMU"; break;
            case VINS::InitStatus::FAIL_PARALLAX:
                oss << "STA: FAIL_PARA"; break;
            case VINS::InitStatus::FAIL_RELATIVE:
                oss << "STA: FAIL_RELA"; break;
            case VINS::InitStatus::FAIL_SFM:
                oss << "STA: FAIL_SFM"; break;
            case VINS::InitStatus::FAIL_PNP:
                oss << "STA: FAIL_PNP"; break;
            case VINS::InitStatus::FAIL_ALIGN:
                oss << "STA: FAIL_ALIGN"; break;
            case VINS::InitStatus::FAIL_CHECK:
                oss << "STA: FAIL_COST"; break;
            case VINS::InitStatus::SUCC:
                oss << "STA: SUCC!"; break;
            default:
                oss << "STA: "; break;
        }
        tvXText = oss.str();
        oss.str(""); oss.clear();
        
        oss << "FAIL: " << vins.fail_times << " times";
        tvYText = oss.str();
        oss.str(""); oss.clear();

        oss << "PARALLAX: " << vins.parallax_num_view;
        tvZText = oss.str();
        oss.str(""); oss.clear();
        
        oss << "Initializing: "<< vins.initProgress;
        tvTotalText = oss.str();
        oss.str(""); oss.clear();

        initImageVisible = true;
    }   
    else {   
        if (finish_init == false) {
            initImageVisible = false;
            LOGD("VINS SOLVER INIT FINISHED.");
            start_show = true;
            finish_init = true;
        }
        
        float x_view = (float)vins.correct_Ps[frame_cnt][0];
        float y_view = (float)vins.correct_Ps[frame_cnt][1];
        float z_view = (float)vins.correct_Ps[frame_cnt][2];
        if (x_view_last == -5000) {
            x_view_last = x_view;
            y_view_last = y_view;
            z_view_last = z_view;
        }
        total_odom += sqrt(pow((x_view - x_view_last), 2) +
        pow((y_view - y_view_last), 2) +
        pow((z_view - z_view_last), 2));
        x_view_last = x_view;
        y_view_last = y_view;
        z_view_last = z_view;
    
        oss << "X: "<< x_view; tvXText = oss.str();
        oss.str(""); oss.clear();
    
        oss << "TOTAL: "<< total_odom;
       tvTotalText = oss.str();
        oss.str(""); oss.clear();
    
        oss << "Y: "<< y_view;
        tvYText = oss.str();
        oss.str(""); oss.clear();

        oss << "Z: "<< z_view;
        tvZText = oss.str();
        oss.str(""); oss.clear();
    }
    oss << "BUF: "<< waiting_lists;
    tvBufText = oss.str();
    oss.str(""); oss.clear();

    if (loop_old_index != -1) {
        oss << "LOOP with: "<< loop_old_index;
        tvLoopText = oss.str();
        oss.str(""); oss.clear();
    }
    oss << "FEATURE: "<< vins.feature_num;
    tvFeatureText = oss.str();
    oss.str(""); oss.clear();
    
    viewUpdateMutex.unlock();
}

void VINSController::switchUI(bool isChecked) {
    if(isChecked) {

        printf("show AR\n");
        ui_main = true;
        box_in_AR = true;
        USE_PNP = true;
        imageCacheEnabled = cameraMode && !USE_PNP;
    } else {

        ui_main = false;
        if (box_in_AR)
            box_in_trajectory = true;
        USE_PNP = false;
        imageCacheEnabled = cameraMode && !USE_PNP;
        printf("show VINS\n");
    }
}

void VINSController::loopButtonPressed(bool isChecked) {
    if(isChecked){
        LOOP_CLOSURE = true;
        LOGI("Loop Closure enabled");
    }
    else{
        LOOP_CLOSURE = false;
        LOGI("Loop Closure disabled");
    }
}

NSTimeInterval VINSController::getLateast_imu_time() const { return lateast_imu_time; }


